#!/usr/bin/env python
#
# Data Quality Checker also known as DQC
#
# Read the README.md for more detailed information
#
#       Barcelona Supercomputing Center - Centro Nacional de Supercomputacion
#       Earth Sciences Department - Computational Earth Sciences
#     
#     Authors: 
#       Joan Sala Calero (joan.sala@bsc.es)
#       Jesus Pena-Izquierdo (jesus.pena@bsc.es)
#       Ivan Jose Cernicharo Ortiz (ivan.cernicharo@bsc.es)

import sys
import logging
import os
import re
import subprocess
import numpy as np
import xarray
import cfgrib
import numpy as np
import pandas as pd
from scipy.stats import kurtosis, skew

# Local imports
from utils.utils_files import *

"""
Class: Valid ranges checks
---
- Checks data to detect outliers
"""
class valid_ranges_test:

    def __init__(self, logdir, path, pattern, varname, vmin, vmax, kstd=10):
        self.testid = 'test01-standard-compliance'
        self.logdir = logdir 
        self.path = path
        self.files = dir_or_file(path, pattern)
        self.varname = varname
        self.stat_columns = ['Min', 'Max', 'Mean', 'Median', 'Kurt', 'Skew', 'Std', 'Mad', 'MC', 'Q1', 'Q3', 'IQD']
        # Number of standard deviations
        self.kstd = kstd
        # Automatic versus computed valid ranges
        try:
            self.valid_min = float(vmin)
        except:
            self.valid_min = None
        try:
            self.valid_max = float(vmax)
        except:
            self.valid_max = None
    
    """
    Producing a list of statistical estimators
    """
    def get_stats(self, data):
        
        # Data to be a flatten numpy array
        Min = np.min(data)
        Max = np.max(data)
        Mean = np.mean(data)
        Median = np.median(data)
        Std = np.std(data)

        # MC = iter_medcouple(data, ni = 1000)
        MC = np.nan #skew(data)
        Q1 = np.quantile(data,0.25)
        Q3 = np.quantile(data,0.75)
        IQD = Q3 - Q1

        # Mad
        diff = np.abs(data - Median)
        Mad = np.median(diff)

        # Kurtosis and Skewness
        Kurt = kurtosis(data)
        Skew = skew(data)
        
        return [Min, Max, Mean, Median, Kurt, Skew, Std, Mad, MC, Q1, Q3, IQD]

    """
    Loading DATA depending on fileformat | All Timesteps | Filter out NaN
    """
    def load_datafile(self, filepath):

        # NetCDF: Read data into Xarray
        if filepath.endswith('.nc'):
            data = xarray.open_dataset(filepath)            
            arr = data[self.varname].isel(time=slice(0,-1)).values
            # Remove NaN values
            arr = arr[~np.isnan(arr)].tolist()

        # GRIB: Use cfgrib package in combination with Xarray
        if filepath.endswith('.grib'):
            data = cfgrib.open_datasets(filepath)                    
            # Get a specific variable (grib files might contain several)
            arr = []
            for d in data:
                if self.varname in d:
                   # Slice all timesteps
                   d3 = d[self.varname].isel(time=slice(0,-1)).values        
                   # Remove NaN values
                   arr += d3[~np.isnan(d3)].tolist()

        return arr

    """
    Main function
    """
    def check_valid_ranges(self):
       
        # Progress init
        i = 0
        n = len(self.files)
        print()

        # Store stats information per file       
        df = pd.DataFrame(index=self.files, columns=self.stat_columns)

        # In a CSV file    
        csv_fname = re.sub("([0-9]{8})", "stats.csv", os.path.splitext(self.files[0])[0]) # YYYYMMDD
        logging.info('Saving stats in {}'.format(csv_fname))
        print('INFO: Saving stats in {}'.format(csv_fname))
        
        # Gather statistics per file
        for f in self.files:

            # Progress update
            i+=1
            print_progress(i, n)              
            logging.info('Checking {}'.format(f))       

            # Load data from file
            data = self.load_datafile(f)

            # Gather statistics            
            df.loc[f] = self.get_stats(data)

        # Save stats for future use
        df.to_csv(csv_fname)

        # Valid range estimation (averaged since we cannot load all data at once)               
        if self.valid_min is None:
            self.valid_min = df['Mean'].mean() - self.kstd * df['Std'].mean()
        if self.valid_max is None:
            self.valid_max = df['Mean'].mean() + self.kstd * df['Std'].mean()

        # Check if any file is outside the limits
        res = True
        for fname, row in df.iterrows():
            if row['Min'] < self.valid_min or row['Max'] > self.valid_max:
                logging.error('{} contains outliers'.format(fname))
                logging.error('[ {} > {} ] vs [ {} > {} ]'.format(self.valid_min, self.valid_max, row['Min'], row['Max']))
                res = False
                break

        # Final message
        if res:
            logging.info('FILES ARE INSIDE VALID RANGES [ FROM {} TO  {} ]'.format(self.valid_min, self.valid_max))
            print('INFO: FILES ARE INSIDE VALID RANGES [ FROM {} TO  {} ]'.format(self.valid_min, self.valid_max))
        else:
            logging.error('FILES CONTAIN UNREALISTIC VALUES OUTSIDE OF VALID RANGES [ FROM {} TO  {} ]'.format(self.valid_min, self.valid_max))
            print('ERROR: FILES CONTAIN UNREALISTIC VALUES OUTSIDE OF VALID RANGES [ FROM {} TO  {} ]'.format(self.valid_min, self.valid_max))

        return res

    def main(self):
        
        # Initiate logging
        initiate_logging(self.logdir, self.path, self.testid)

        # Check valid ranges
        self.check_valid_ranges()
    
    
if __name__ == '__main__':
    sys.exit()
